;compiler: avrasm
	.include "m32def.inc"
	.include "macros.asm"
	
	;XTAL freq
	.equ	XTAL=16000000
	
	;PS/2 pins (clk, data)
	.equ	PS2ddr=ddrb
	.equ	PS2pins=pinb
	.equ	PS2c=0
	.equ	PS2d=1
	

	.dseg
PS2timeoutT:	.byte	1
MouseXpos:	.byte	2
MouseYpos:	.byte	2
MouseZpos:	.byte	2
MouseID:	.byte	1
RXbyte:	.byte	1
PS2state:	.byte	1
MouseType:	.byte	1
	.cseg
	
	.org	0x0000
	rjmp	Init
	.org	0x0014
	rjmp	INT_PS2timeout
	.org	0x001A
	rjmp	INTusartRXC
	

;************************************************************************************************************************
;	USART RXC INT
;************************************************************************************************************************
INTusartRXC:	pushw	R16
	in	R17,sreg
	;
	in	R16,udr	;read recieved byte
	sts	RXbyte,R16	;store it
	lds	R16,PS2state
	cpi	R16,'r'
	breq	INTusartRXCs
	;send mouse init status if mouse not initialized
	rcall	COMtxByte
	stsi	RXbyte,0x00
INTusartRXCs:	;
	out	sreg,R17
	popw	R16
	reti

;************************************************************************************************************************
;	PS/2 timeout INT
;************************************************************************************************************************
INT_PS2timeout:	pushw	R16
	in	R17,sreg
	;
	lds	R16,PS2timeoutT
	dec	R16	;decrement timeout counter
	sts	PS2timeoutT,R16
	brne	INT_PS2timeoutS
	;stop timer if timeout, T=1
	outi	tccr0,0
	outi	tifr,0xFF
	ori	R17,1<<6	;"set"
INT_PS2timeoutS:	;
	out	sreg,R17	
	popw	R16
	reti
;------------------------------------------------------------------------------------------------------------------------
;	initialize PS/2 timeout timer 
;------------------------------------------------------------------------------------------------------------------------
PS2timeoutS:	in	R0,sreg
	cli
	nop
	sts	PS2timeoutT,R16
	outi	tccr0,0
	outi	tcnt0,0
	outi	tccr0,(1<<ctc0)|(1<<cs02)|(1<<cs00)
	outi	tifr,0xFF
	out	sreg,R0
	clt
	ret
;------------------------------------------------------------------------------------------------------------------------
;	stop timout timer
;------------------------------------------------------------------------------------------------------------------------	
PS2timeoutP:	in	R0,sreg
	cli
	nop
	push	R16
	outi	tccr0,0
	outi	tifr,0xFF
	pop	R16
	out	sreg,R0
	ret



	

;************************************************************************************************************************
;	initialization
;************************************************************************************************************************	
Init:	;stack pointer
	outiw	sp,ramend
	
	;timer 0 for PS/2 recieve timeout	
	outi	ocr0,XTAL/(1024*100)-1
	outi	timsk,(1<<ocie0)
	
	;usart, rxen, txen, rxcie
	outi	ucsra,(1<<u2x)
	outi	ucsrb,(1<<rxen)|(1<<txen)|(1<<rxcie)
	outiw	ubrr,16	;115200bd for XTAL=16MHz
	
	;wait 100ms
	ldi	R16,10
	rcall	DelayVar
		
	stsi	RXbyte,0x00	;no PC command
	stsi	PS2state,'n'	;no PS/2 device detected
	sei
	
	
	cbi	PS2ddr,PS2c	;allow device clocking
		
PS2mouseReset:	ldi	R16,0xFF	;reset cmd
	rcall	PS2writeByte	;write cmd
	rjts	PS2mouseErr	;timeout?
	rcall	PS2readByte	;read ack byte
	rcall	PS2readByte	;read selftest byte
	rcall	PS2readByte	;read ID byte
	
	stsi	PS2state,'d'	;status: PS/2 device detected
	
	ldi	R16,70
	rcall	DelayVar	;wait
	
	ldi	R16,0xFF	;reset cmd again
	rcall	PS2writeByte	;write cmd
	rjts	PS2mouseErr	;timeout?
	rcall	PS2readByte	;read ack byte
	rcall	PS2readByte	;read selftest byte
	rcall	PS2readByte	;read ID byte

	
	ldi	R16,0xF2	;read ID cmd
	rcall	PS2writeByte	;write cmd
	rjts	PS2mouseErr	;timeout?
	rcall	PS2readByte	;read ack byte
	rcall	PS2readByte	;read ID byte
	cpi	R16,0xAB	;keyboard?
	brne	NoKeybrd
	stsi	PS2state,'k'	;status: PS/2 keyboard detected
	ldi	R16,200
	rcall	DelayVar	;wait
	rjmp	PS2mouseReset	;reinit
NoKeybrd:	cpi	R16,0x00	;mouse?
	rjne	PS2mouseReset	;not mouse - reinit
	;standard mouse detected
	
	;try to enable M$ Intelli mouse (set sample rate 200,100,80)
	;set sample rate 200
	ldi	R16,0xF3
	rcall	PS2writeByte
	rjts	PS2mouseErr	;timeout?
	rcall	PS2readByte
	ldi	R16,200
	rcall	PS2writeByte
	rcall	PS2readByte
	
	;set sample rate 100
	ldi	R16,0xF3
	rcall	PS2writeByte
	rjts	PS2mouseErr	;timeout?
	rcall	PS2readByte
	ldi	R16,100
	rcall	PS2writeByte
	rcall	PS2readByte
	
	;set sample rate 80
	ldi	R16,0xF3
	rcall	PS2writeByte
	rjts	PS2mouseErr	;timeout?
	rcall	PS2readByte
	ldi	R16,80
	rcall	PS2writeByte
	rcall	PS2readByte
		
	;read ID again
	ldi	R16,0xF2
	rcall	PS2writeByte
	rjts	PS2mouseErr	;timeout?
	rcall	PS2readByte
	rcall	PS2readByte
	sts	MouseID,R16
	ldi	R17,'m'
	cpi	R16,0x03	;M$ Intelli mouse?
	brne	PC+2
	ldi	R17,'i'	;hell yeah
	sts	PS2state,R17	;store status
	sts	MouseType,R17	;store mouse type
	ldi	R16,200
	rcall	DelayVar	;wait
		
	;set remote mode
	ldi	R16,0xF0
	rcall	PS2writeByte
	rjts	PS2mouseErr	;timeout?
	rcall	PS2readByte
	
	;set resolution
	ldi	R16,0xF8
	rcall	PS2writeByte
	rjts	PS2mouseErr	;timeout?
	rcall	PS2readByte
	ldi	R16,0x03	;8/mm
	rcall	PS2writeByte
	rcall	PS2readByte

	stsiw	MouseXpos,0	;dx=0
	stsiw	MouseYpos,0	;dy=0
	stsiw	MouseZpos,0	;dz=0
	
Loop:	stsi	PS2state,'r'	;status: remote mode initialized 

	ldi	R16,0xEB	;read data cmd
	rcall	PS2writeByte	;write cmd
	rjts	PS2mouseErr	;timeout?
	rcall	PS2readByte	;read ack byte
	rjts	PS2mouseErr	;timeout?
	rcall	PS2readByte	;read byte 0
	rjts	PS2mouseErr	;timeout?
	mov	R4,R16
	rcall	PS2readByte	;read byte 1
	rjts	PS2mouseErr	;timeout?
	mov	R5,R16
	rcall	PS2readByte	;read byte 2
	rjts	PS2mouseErr	;timeout?
	mov	R6,R16
	lds	R16,MouseID
	cpi	R16,0x03
	brne	LoopNoMSI
	rcall	PS2readByte	;read byte 3 if ID=0x03
	rjts	PS2mouseErr	;timeout?
	mov	R7,R16	
LoopNoMSI:		
	;byte 0 bit flags
	.equ	ps2bSignX=4
	.equ	ps2bSignY=5
	.equ	ps2bBtnR=1
	.equ	ps2bBtnM=2
	.equ	ps2bBtnL=0

	;local_dx+=mouse_dx
	ldsw	R16,MouseXpos
	clr	R18
	sbrc	R4,ps2bSignX
	dec	R18
	add	R16,R5
	adc	R17,R18	;add dx
	stsw	MouseXpos,R16
	
	;local_dy+=mouse_dy
	ldsw	R16,MouseYpos
	clr	R18
	sbrc	R4,ps2bSignY
	dec	R18
	add	R16,R6
	adc	R17,R18	;add dy
	stsw	MouseYpos,R16
	
	lds	R16,MouseID
	cpi	R16,0x03
	brne	SkipZ
	;M$ Intelli mouse
	
	;local_dz+=mouse_dz	
	ldsw	R16,MouseZpos
	clr	R18
	sbrc	R7,3
	dec	R18
	add	R16,R7
	adc	R17,R18
	stsw	MouseZpos,R16
SkipZ:	
	;now test for 's' cmd
	cli
	nop
	lds	R16,RXbyte
	cpi	R16,0x00
	rjeq	PCcommandNo
	;some command
	cpi	R16,'s'	;return status?
	rjne	PCcommandNo
	;'s' - return status
	ldi	R16,'a'
	rcall	COMtxByte	;return 'a'
	lds	R16,MouseType
	rcall	COMtxByte	;return 'm'/'i'
	
	;return dx
	lds	R16,MouseXpos+0
	rcall	COMtxByte
	lds	R16,MouseXpos+1
	rcall	COMtxByte
	;return dy
	lds	R16,MouseYpos+0
	rcall	COMtxByte
	lds	R16,MouseYpos+1
	rcall	COMtxByte
	;return dz
	lds	R16,MouseZpos+0
	rcall	COMtxByte
	lds	R16,MouseZpos+1
	rcall	COMtxByte
	;return mouse packet byte 0	
	mov	R16,R4
	rcall	COMtxByte
	
	stsiw	MouseXpos,0	;dx=0
	stsiw	MouseYpos,0	;dy=0
	stsiw	MouseZpos,0	;dz=0
	
PCcommandNo:	stsi	RXbyte,0x00	;no cmd
	sei
		
	rjmp	Loop	;repeat
	
	;mouse timeout/error
PS2mouseErr:	stsi	PS2state,'n'	;status: no PS/2 device detected
	rjmp	PS2mouseReset
	

	
;************************************************************************************************************************
;	PS/2 routines
;************************************************************************************************************************	
;------------------------------------------------------------------------------------------------------------------------
;	PS/2 read byte into R16, T=1 if timeout 
;------------------------------------------------------------------------------------------------------------------------
PS2readByte:	ldi	R16,50
	rcall	PS2timeoutS	;init timeout timer
	brts	PS2readByteTO
	sbic	PS2pins,PS2c
	rjmp	PC-2	;wait for device clock low
	ldi	R16,50
	rcall	PS2timeoutS	;init timeout timer
	brts	PS2readByteTO
	sbis	PS2pins,PS2c
	rjmp	PC-2	;wait for clock high
	;read data bits
	ldi	R17,8	;8 data bits
PS2readByteL:	brts	PS2readByteTO
	sbic	PS2pins,PS2c
	rjmp	PC-2	;wait for clock low
	rcall	PS2delay10	;wait
	rjis	PS2pins,PS2d,PS2readByteDH
	clc		;data low
	rjmp	PS2readByteDL	
PS2readByteDH:	sec		;data high
PS2readByteDL:	ror	R16	;store data bit
	brts	PS2readByteTO
	sbis	PS2pins,PS2c
	rjmp	PC-2	;wait for clock high
	dbnz	R17,PS2readByteL
	;parity bit (ignored)
	brts	PS2readByteTO
	sbic	PS2pins,PS2c
	rjmp	PC-2	;wait for device clock low
	brts	PS2readByteTO
	sbis	PS2pins,PS2c
	rjmp	PC-2	;wait for clock high
	;stop bit (ignored)
	brts	PS2readByteTO
	sbic	PS2pins,PS2c
	rjmp	PC-2	;wait for device clock low
	rcall	PS2delay10
	brts	PS2readByteTO
	sbis	PS2pins,PS2c
	rjmp	PC-2	;wait for clock high
	clt
PS2readByteTO:	rcall	PS2delay100	;wait
	rcall	PS2timeoutP	;stop timeout timer
	ret
	
;------------------------------------------------------------------------------------------------------------------------
;	write R16 byte, T=1 if timeout
;------------------------------------------------------------------------------------------------------------------------	
PS2writeByte:	;get parity bit
	mov	R1,R16
	swap	R1
	eor	R1,R16
	mov	R17,R1
	lsr	R1
	lsr	R1
	eor	R1,R17
	mov	R17,R1
	lsr	R1
	eor	R1,R17
	;
	ldi	R17,8	;8 bits to send
	sbi	PS2ddr,PS2c	;clock low
	rcall	PS2delay100	;wait 100us
	sbi	PS2ddr,PS2d	;data low
	rcall	PS2delay10	;wait 10us
	cbi	PS2ddr,PS2c	;clock high
	rcall	PS2delay10	;wait 10us
	;device should start clocking
	push	R16
	ldi	R16,70
	rcall	PS2timeoutS	;init timeout timer
	pop	R16
PS2writeByteL:	brts	PS2writeByteTO
	sbic	PS2pins,PS2c
	rjmp	PC-2	;wait for clock low
	rcall	PS2delay10	;wait 10us
	;transfer start
	rjrs	R16,0,PS2writeByteDH	;low/high bit?
	sbi	PS2ddr,PS2d	;data low
	rjmp	PS2writeByteDL
PS2writeByteDH:	cbi	PS2ddr,PS2d	;data high
PS2writeByteDL:	lsr	R16	;next bit
	brts	PS2writeByteTO	
	sbis	PS2pins,PS2c
	rjmp	PC-2	;wait for clock high
	dbnz	R17,PS2writeByteL
	;send parity bit
	brts	PS2writeByteTO
	sbic	PS2pins,PS2c
	rjmp	PC-2	;wait for clock low
	rcall	PS2delay10	;wait 10us
	rjrc	R1,0,PS2writeBytePH	;parity low/high?
	sbi	PS2ddr,PS2d	;data low	
	rjmp	PS2writeBytePL
PS2writeBytePH:	cbi	PS2ddr,PS2d	;data high
PS2writeBytePL:	brts	PS2writeByteTO
	sbis	PS2pins,PS2c
	rjmp	PC-2	;wait for clock high
	brts	PS2writeByteTO
	sbic	PS2pins,PS2c
	rjmp	PC-2	;wait for clock low
	;stop bit
	rcall	PS2delay10	;wait
	cbi	PS2ddr,PS2d	;data high
	;ACK
	brts	PS2writeByteTO
	sbic	PS2pins,PS2d
	rjmp	PC-2	;wait for data low
	brts	PS2writeByteTO
	sbic	PS2pins,PS2c
	rjmp	PC-2	;wait for clock low
	brts	PS2writeByteTO
	sbis	PS2pins,PS2c
	rjmp	PC-2	;wait for clock high
	brts	PS2writeByteTO
	sbis	PS2pins,PS2d
	rjmp	PC-2	;wait for data high
	;done
	clt
	rcall	PS2timeoutP
	ret
	;timeout exit
PS2writeByteTO:	cbi	PS2ddr,PS2d	;data high
	rcall	PS2delay100	;wat 100us
	cbi	PS2ddr,PS2c	;clk high
	rcall	PS2timeoutP	;stop timeout timer
	ret

;------------------------------------------------------------------------------------------------------------------------
;	wait 10us
;------------------------------------------------------------------------------------------------------------------------
PS2delay10:	push	R16
	ldi	R16,XTAL/(3*100000)-3
PS2delay10L:	dbnz	R16,PS2delay10L
	pop	R16
	ret

;------------------------------------------------------------------------------------------------------------------------
;	wait 100us
;------------------------------------------------------------------------------------------------------------------------	
PS2delay100:	push	R16
	ldi	R16,XTAL/(8*10000)-2
PS2delay100L:	adiw	R30,0
	adiw	R30,0
	nop
	dbnz	R16,PS2delay100L
	pop	R16
	ret

	


;************************************************************************************************************************
;	USART routines
;************************************************************************************************************************	
;------------------------------------------------------------------------------------------------------------------------
;	send R16 byte
;------------------------------------------------------------------------------------------------------------------------
COMtxByte:	sbis	ucsra,udre
	rjmp	PC-1
	out	udr,R16
	ret

	

;************************************************************************************************************************
;	delay routines
;************************************************************************************************************************
;------------------------------------------------------------------------------------------------------------------------
;	variable delay (R16 * 10ms)
;------------------------------------------------------------------------------------------------------------------------
DelayVar:	pushw	R18
DelayVar_3:	ldi	R19,250
DelayVar_2:	ldi	R18,XTAL/74000
DelayVar_1:	dbnz	R18,DelayVar_1
	dbnz	R19,DelayVar_2
	dbnz	R16,DelayVar_3
	popw	R18
	ret